package com.dsk.flink.dsc.sync;

import cn.hutool.core.collection.CollUtil;
import cn.hutool.core.util.StrUtil;
import com.alibaba.fastjson.JSONObject;
import com.dsk.flink.dsc.common.function.AsyncDamengDataTransferFunction;
import com.dsk.flink.dsc.common.function.CanalMapToTsGroupFunction;
import com.dsk.flink.dsc.common.function.GroupTsProcessWindowFunction;
import com.dsk.flink.dsc.common.sink.DamengDataTransferSink;
import com.dsk.flink.dsc.utils.EnvProperties;
import com.dsk.flink.dsc.utils.EnvPropertiesUtil;
import com.dsk.flink.dsc.utils.EtlUtils;
import io.tidb.bigdata.tidb.JdbcConnectionProviderFactory;
import lombok.extern.slf4j.Slf4j;
import org.apache.flink.api.common.functions.FilterFunction;
import org.apache.flink.api.common.restartstrategy.RestartStrategies;
import org.apache.flink.api.common.serialization.SimpleStringSchema;
import org.apache.flink.api.java.tuple.Tuple3;
import org.apache.flink.api.java.utils.ParameterTool;
import org.apache.flink.streaming.api.CheckpointingMode;
import org.apache.flink.streaming.api.datastream.AsyncDataStream;
import org.apache.flink.streaming.api.datastream.SingleOutputStreamOperator;
import org.apache.flink.streaming.api.environment.CheckpointConfig;
import org.apache.flink.streaming.api.environment.StreamExecutionEnvironment;
import org.apache.flink.streaming.api.functions.windowing.ProcessWindowFunction;
import org.apache.flink.streaming.api.windowing.assigners.TumblingProcessingTimeWindows;
import org.apache.flink.streaming.api.windowing.time.Time;
import org.apache.flink.streaming.api.windowing.windows.TimeWindow;
import org.apache.flink.streaming.connectors.kafka.FlinkKafkaConsumer;
import org.apache.flink.util.Collector;

import java.time.LocalDateTime;
import java.time.ZoneId;
import java.util.Comparator;
import java.util.List;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;

/**
 * @author shezaixing
 * @date 2023/12/5 14:44
 * @description 同步至客户目标数据源任务
 */
@Slf4j
public class SyncCustomerDataSource {

    public static void main(String[] args) throws Exception {

        StreamExecutionEnvironment env = StreamExecutionEnvironment.getExecutionEnvironment();
        env.setRestartStrategy(RestartStrategies.fixedDelayRestart(3, 30000));
        env.enableCheckpointing(60000);
        env.getCheckpointConfig().setCheckpointingMode(CheckpointingMode.EXACTLY_ONCE);
        env.getCheckpointConfig().setMinPauseBetweenCheckpoints(60000);
        env.getCheckpointConfig().setCheckpointTimeout(7200000);
        env.getCheckpointConfig().setMaxConcurrentCheckpoints(1);
        env.getCheckpointConfig().setExternalizedCheckpointCleanup(CheckpointConfig.ExternalizedCheckpointCleanup.RETAIN_ON_CANCELLATION);
        env.getCheckpointConfig().setTolerableCheckpointFailureNumber(3);

        //获取用户自己的配置信息
        ParameterTool parameterTool = ParameterTool.fromArgs(args);
        String offsetTimestamp = parameterTool.get("offsetTimestamp");
        String propertiesPath = parameterTool.get("propertiesPath");
        EnvProperties envProps = EnvPropertiesUtil.getPropertiesFromArgsPath(propertiesPath);
        envProps.put("providerImpl",
                JdbcConnectionProviderFactory.HikariDataSourceJdbcConnectionProvider.class.getName());
        System.out.println("读取到的配置文件：-> " + envProps.toString());
        System.out.println("读取到的数据连接配置：->" + String.format(envProps.getDb_url(), envProps.getDb_host(),
                envProps.getDb_port(), envProps.getDb_database()));
        System.out.println("获取到的kafka消费组：->" + EtlUtils.getKafkaGroup(envProps));
        //TODO 到时需要改这里，改成正式的消费组
        FlinkKafkaConsumer<String> kafkaConsumer = new FlinkKafkaConsumer<String>(envProps.getKafka_topic(),
                new SimpleStringSchema(), EtlUtils.getKafkaConfig(envProps.getKafka_brokers(),
                EtlUtils.getKafkaGroup(envProps), envProps.getKafka_username(), envProps.getKafka_password()));
        //System.out.println(envProps.getKafka_topic());
        long defaultOffset =
                LocalDateTime.now().minusMinutes(10).atZone(ZoneId.systemDefault()).toInstant().toEpochMilli();
        kafkaConsumer.setStartFromTimestamp(defaultOffset);
        //kafkaConsumer.setStartFromLatest();
        //偏移量
        if (StrUtil.isNotBlank(offsetTimestamp)) {
            kafkaConsumer.setStartFromTimestamp(Long.parseLong(offsetTimestamp));
        }

        SingleOutputStreamOperator<String> kafkaSource = env.addSource(kafkaConsumer)
                .setParallelism(1)
                .name("kafka-source")
                .uid("kafka-source");

        //kafkaSource.print("kafaka stream");

        SingleOutputStreamOperator<JSONObject> canalJsonStream = kafkaSource.map(JSONObject::parseObject)
                .filter(new FilterFunction<JSONObject>() {
                    @Override
                    public boolean filter(JSONObject value) throws Exception {

                        return !value.getBoolean("isDdl") && !"TIDB_WATERMARK".equals(value.getString("type"));
                    }
                })
                .name("canalJsonStream")
                .uid("canalJsonStream");

        //canalJsonStream.print("canal stream");

        SingleOutputStreamOperator<Tuple3<JSONObject, String, Long>> tsGroupStream = canalJsonStream.map(new CanalMapToTsGroupFunction());

        SingleOutputStreamOperator<JSONObject> process = tsGroupStream.keyBy(x -> x.f2)
                .window(TumblingProcessingTimeWindows.of(Time.seconds(3)))
                .process(new GroupTsProcessWindowFunction());

//        SingleOutputStreamOperator<String> sqlResultStream = AsyncDataStream.orderedWait(canalJsonStream, new
//        AsyncMysqlDataTransferFunction(envProps), 1200L, TimeUnit.SECONDS, 20)
//                .filter(new FilterFunction<String>() {
//                    @Override
//                    public boolean filter(String value) throws Exception {
//                        return StrUtil.isNotBlank(value) && !"err".equals(value);
//                    }
//                })
//                .name("sqlResultStream")
//                .uid("sqlResultStream");
//
//        //sqlResultStream.print("sql result");
//
//        sqlResultStream.addSink(new MysqlDataTransferSink(envProps)).name("sqlSinkStream").uid("sqlSinkStream");

        SingleOutputStreamOperator<Tuple3<String, String, Long>> sqlResultStream1 =
                AsyncDataStream.orderedWait(process, new AsyncDamengDataTransferFunction(envProps), 1200L, TimeUnit.SECONDS, 20)
                .filter(new FilterFunction<Tuple3<String, String, Long>>() {
                    @Override
                    public boolean filter(Tuple3<String, String, Long> value) throws Exception {
                        return StrUtil.isNotBlank(value.f0) && !"err".equals(value.f0);
                    }
                })
                .name("sqlResultStream")
                .uid("sqlResultStream");

        SingleOutputStreamOperator<String> groupWindowSqlResultStream = sqlResultStream1.keyBy(value -> value.f1)
                .window(TumblingProcessingTimeWindows.of(Time.seconds(3)))
                .process(new ProcessWindowFunction<Tuple3<String, String, Long>, String, String, TimeWindow>() {
                    @Override
                    public void process(String s, ProcessWindowFunction<Tuple3<String, String, Long>, String, String,
                            TimeWindow>.Context context, Iterable<Tuple3<String, String, Long>> elements,
                                        Collector<String> out) throws Exception {
                        List<Tuple3<String, String, Long>> list = CollUtil.list(false, elements);
                        if ("dsc_cdc_log".equals(list.get(0).f1)) {
                            list = list.stream().sorted(Comparator.comparing(x -> x.f2, Comparator.reverseOrder())).collect(Collectors.toList());
                            list.forEach(x -> {
                                out.collect(x.f0);
                            });
                            return;
                        }
                        Tuple3<String, String, Long> maxTsElement =
                                list.stream().max(Comparator.comparing(x -> x.f2)).get();
                        out.collect(maxTsElement.f0);
                    }
                })
                .name("groupWindowSqlResultStream")
                .uid("groupWindowSqlResultStream");

        //groupWindowSqlResultStream.print("sql result");

        groupWindowSqlResultStream.addSink(new DamengDataTransferSink(envProps)).name("sqlSinkStream").uid("sqlSinkStream");

        env.execute();
    }
}
